from typing import Optional, Literal, TypeAlias

from packaging.version import Version
from pyvista.plotting import Plotter
from streamlit.components.v2 import component
from streamlit.version import STREAMLIT_VERSION_STRING

from .dataview import dataview
from .panel_backend import PanelVTKKwargs, _panel_html
from .trame_backend import _trame_html

# Type definitions
WindowDimension: TypeAlias = int | Literal["stretch", "content"]

# Check Streamlit version supports components.v2
if Version(STREAMLIT_VERSION_STRING) < Version("1.51.0"):
    raise ImportError(
        f"Streamlit version {STREAMLIT_VERSION_STRING} should be >= 1.51"
        "Pin stpyvista<=0.1.4 for components.v1 compatibility"
    )

# Declare component to create callable
_html = (
    '<div id="stpyvistadiv">'
    '<iframe id="stpyvistaframe" frameborder="0" allowfullscreen allowtransparency referrerpolicy="same-origin">'
    "</iframe></div>"
)

_stpv_component = component(
    "stpyvista.simple",
    html=_html,
    js="v2/stpyvista.js",
    css="v2/style.css",
    isolate_styles=True,
)

# Expose this to streamlit user
def stpyvista(
    plotter: Plotter,
    backend: Literal["panel", "trame"] = "trame",
    backend_kwargs: Optional[PanelVTKKwargs] = None,
    width: WindowDimension = "stretch",
    key: Optional[str] = None,
) -> None:
    """
    Renders an interactive Pyvista Plotter in streamlit using either panel or trame backends.

    Parameters
    ----------
    plotter: pv.Plotter
        Pyvista plotter object to render.

    backend: Literal["panel", "trame"] = "trame",
        Supported backends are `panel` or `trame`. Each requires separate dependencies.

    backend_kwargs : Optional[PanelVTKKwargs | dict] = None
        Optional keyword parameters to pass to `pn.panel()`. Check the `PanelVTKKwargs`
        documentation for details. Here is a list with some useful ones:

        axes: PanelAxesConfig
            Parameters of the axes to construct in the 3D view.

        orientation_widget : bool
            Show the xyz axis indicator

        interactive_orientation_widget: bool
            Show and interactive xyz axis indicator

        See the example guide at https://stpyvista.streamlit.app/?gallery=axes

    width =: WindowDimension = "stretch"
        The width of the stpyvista plotter. If set to "stretch" (default), the width matches the parent container. If 
        set to "content", the width matches the `plotter.window_size`. If an integer is passed, it is set as the
        the width in pixels. 
    
    key: Optional[str] = None
        An optional key that uniquely identifies this component. If this is None, and the component's arguments are 
        changed, the component will be re-mounted in the Streamlit frontend and lose its current state.
    """

    if not isinstance(plotter, Plotter):
        raise TypeError(f"{plotter} is not a `pyvista.Plotter` instance.")

    if backend_kwargs is None:
        backend_kwargs = {}

    # Determine width
    use_container_width = width == "stretch"
    plotter_width, height = plotter.window_size

    if not use_container_width:
        if not isinstance(width, int):
            width = plotter_width

    if backend == "panel":
        iframe = _panel_html(plotter, use_container_width=use_container_width, **backend_kwargs)
    elif backend == "trame":
        iframe = _trame_html(plotter, **backend_kwargs)
    else:
        raise ValueError("backend value not supporter")

    data = {
        "_html": iframe,
        "backend": backend,
        "height": height,
        "width": width,
        "use_container_width": use_container_width,
    }

    component_value = _stpv_component(key=key, data=data)
    return component_value


__all__ = ["dataview", "stpyvista"]



from multiprocessing import Process, Queue
from pyvista.plotting import Plotter


def _trame_html(plotter: Plotter) -> str:
    if not isinstance(plotter, Plotter):
        raise TypeError(f"{plotter} is not a `pyvista.Plotter` instance.")

    def _export_html(queue: Queue, plotter: Plotter):
        queue.put(plotter.export_html(filename=None))

    queue = Queue(maxsize=1)
    process = Process(target=_export_html, args=(queue, plotter))

    process.start()
    html_plotter = queue.get().read()
    process.join()

    return html_plotter



from io import StringIO
from typing import TypedDict

import panel as pn
from pyvista import Plotter

try:
    # Available from Python 3.11
    from typing import Required, NotRequired

except ImportError:
    # Fallback for older Python versions.
    #   typing_extensions is not a dependency of stpyvista but it
    #   is a dependency of streamlit, so it should be available.
    from typing_extensions import Required, NotRequired


pn.extension("vtk", sizing_mode="stretch_both")


class PanelTicker(TypedDict):
    """
    A `PanelTicker` is a dictionary that contains:
    - **ticks** : `list[float]`, required.
        Positions in the scene coordinates of the corresponding axis' ticks.
    - **labels** : `list[str]`, optional.
        Label displayed respectively to the **ticks** positions.
        If **labels** are not defined, they are inferred from the **ticks** list.
    """

    ticks: list[float]
    labels: NotRequired[list[str]]


class PanelAxesConfig(TypedDict, total=False):
    """
    A `PanelAxesConfig` is a dictionary containing the parameters of the
    axes to construct in the 3D view. It **must contain**:
    - **xticker**: `PanelTicker`, required
    - **yticker**: `PanelTicker`, required
    - **zticker**: `PanelTicker`, required

    Other optional parameters for `PanelAxesConfig` are:
    - **digits** : `int`
        number of decimal digits when `ticks` are converted to `labels`.
    - **fontsize** : `int`
        size in pts of the ticks labels.
    - **show_grid** : `bool`, default is True
        if true the axes grid is visible.
    - **grid_opacity** : `float`, between 0-1
        defines the grid opacity.
    - **axes_opacity** : `float`, between 0-1
        defines the axes lines opacity.
    """

    xticker: Required[PanelTicker]
    yticker: Required[PanelTicker]
    zticker: Required[PanelTicker]
    origin: tuple[float, float, float]
    fontsize: int
    show_grid: bool
    grid_opacity: float
    axes_opacity: float
    digits: int


class PanelVTKKwargs(TypedDict, total=False):
    """
    A `PanelVTKKwargs` is a dictionary containing the parameters of the
    vtk pane to construct in the 3D view. Optional parameters are:
    - **axes**: `PanelAxesConfig`
        parameters of the axes to construct in the 3D view.
    - **orientation_widget** : `bool`
        show the xyz axis indicator
    - **interactive_orientation_widget** : `bool`
        show and interactive xyz axis indicator
    """

    axes: PanelAxesConfig
    orientation_widget: bool
    interactive_orientation_widget: bool


def _panel_html(plotter: Plotter, **backend_kwargs) -> str:

    use_container_width = backend_kwargs.pop("use_container_width", False)
    width, height = plotter.window_size
    backend_kwargs["height"] = height 
    
    if not use_container_width:
        backend_kwargs["width"] = width
    
    vtk_pane = pn.pane.VTK(plotter.ren_win, **backend_kwargs)

    # Create HTML file
    with StringIO() as model_bytes:
        vtk_pane.save(
            model_bytes,
            title="Running stpyvista.panel_based",
        )
        html_plotter = model_bytes.getvalue()

    return html_plotter



from base64 import b64encode
from pathlib import Path
from typing import Optional
from io import StringIO

import streamlit.components.v1 as components
import panel as pn
from trame.app import get_server
from trame.widgets import (
    vuetify as vuetify,
)
from trame.widgets.vtk import VtkLocalView
from trame.ui.vuetify import SinglePageLayout
from pyvista.plotting import Plotter


# Tell streamlit that there is a component called `experimental_vtkjs`,
# and that the code to display that component is in the "vanilla_vtkjs" folder
frontend_dir = (Path(__file__).parent / "backends/v1/vtkjs_based").absolute()
_stpv_vtkjs_component = components.declare_component("experimental_vtkjs", path=str(frontend_dir))

SERVER_NAME = "stpyvista_server"


async def export_vtksz(plotter: Plotter) -> bytes:
    """Export this plotter as a VTK.js OfflineLocalView file.

    Parameters
    ----------
    plotter : Plotter
        PyVista Plotter object.

    Returns
    -------
    bytes
        The exported plotter view.
    """

    # Get a trame server and launch it
    server = get_server(name=SERVER_NAME, client_type="vue2")
    _, ctrl = server.state, server.controller

    with SinglePageLayout(server) as layout:
        with layout.content:
            view = VtkLocalView(plotter.ren_win)
            ctrl.view_update = view.update

    server.start(
        exec_mode="task",
        host="127.0.0.1",
        port="0",
        open_browser=False,
        show_connection_info=False,
        disable_logging=True,
        timeout=0,
        backend="tornado",
    )

    content = view.export(format=format)
    view.release_resources()

    return content


def stpyvista(vtksz_data: bytes, height: int = 400, key: Optional[str] = None) -> dict:
    """
    Renders an interactive Pyvista Plotter in streamlit.

    Parameters
    ----------
    vtksz_data: bytes
        Data from a vtksz file in zip format.

    Returns
    -------
    dict
        A dictionary with the current Camera view properties.
    """

    base64_str = b64encode(vtksz_data).decode().replace("\n", "")

    component_value = _stpv_vtkjs_component(
        plotter_data=base64_str,
        height=str(height),
        key=key,
        default=0,
    )

    return component_value


def _vtkjs_html(plotter: Plotter, **backend_kwargs) -> str:
    width, height = plotter.window_size
    vtk_pane = pn.pane.VTK(plotter.ren_win, width=width, height=height, **backend_kwargs)

    # Create HTML file
    with StringIO() as model_bytes:
        vtk_pane.save(
            model_bytes,
            title="Running stpyvista",
        )
        html_plotter = model_bytes.getvalue()

    return html_plotter
